from graphsTrees.trees.node import Node
from graphsTrees.exceptions.TreeExceptions import NodeError,NodeKeyError

class Tree:
    
    def __init__(self):
        self._preorder_list = []
        self._inorder_list = []
        self._postorder_list = []
   
    def create_node(self,key):
        try: 
            return Node(int(key))
        except ValueError:
            raise NodeKeyError
    
    def insert(self, node, key):
        if node is None:
            return self.create_node(int(key))
        else:
            node.left = self.insert(node.left, int(key))
        return node
    
    def search_node(self,node,key):
        try: 
            if node is None:
                return "Node {} does not exist".format(int(key))
            if node.key == key:
                return "Node {} exists".format(int(key))
            else:
                return self.search_node(node.left,int(key))
        
        except ValueError:
            raise NodeKeyError
            
    def preorder(self, node):
        try: 
            if node is not None:
                self._preorder_list.append(node.key)
                self.preorder(node.left)
                self.preorder(node.right)
        
        except TypeError:
            raise NodeError
    
    def inorder(self,node):
        try: 
            if node is not None:
                self.inorder(node.left)
                self._inorder_list.append(node.key)
                self.inorder(node.right)
        
        except TypeError:
            raise NodeError
            
    
    def postorder(self, node):
        try: 
            if node is not None: 
                self.postorder(node.left)
                self.postorder(node.right)
                self._postorder_list.append(node.key)
        
        except TypeError:
            raise NodeError    
    
    def get_preorder(self):
        return self._preorder_list
    
    def get_inorder(self):
        return self._inorder_list
    
    def get_postorder(self):
        return self._postorder_list






import numpy as np
import pandas as pd

from graphsTrees.exceptions.EricExceptions import vertexError
from graphsTrees.exceptions.EricExceptions import edgeError
#Input must be of form vertex: integer. edges: list of lists of two elements of vertex set.

# I wanted to implement graphs in a way that wasn't adjacency matrices or adjacency lists
# To show that I didn't plagiarize from teh internet... Because no competent programmer would
# Have ever used this horrible methodology

class Graph:
    
    # This parameter will change first time user runs hasCycles method, or DFS on a cyclic vertex
    __cycles = False #Want these to be hidden
    __connected = None
    
    def __init__(self, vertices, edges):
        try:
            self.vertices = np.arange(vertices)
        except TypeError:
            raise vertexError('Vertices is an integer, determining number of vertices')
        self.edges = edges
        try:
            for edge in edges:
                if len(edge) != 2 or type(edge) != list:
                    raise edgeError('Edges must be two element lists')
                if edge[0] not in self.vertices or edge[1] not in self.vertices:
                    raise edgeError('Edge connects non-existant vertex')
        except TypeError:
            raise edgeError("Edge set must be a list of two element lists")
    
    def addEdge(self, v1, v2):
        if type(v1) != int or type(v2) != int:
            raise vertexError("Vertices are integers")
        elif (v1 in self.vertices and v2 in self.vertices):
            if [v1, v2] not in self.edges and [v2, v1] not in self.edges:
                self.edges.append([v1, v2])
            else:
                return('Edge already in edge set')
        else:
            return('Input valid vertices')
            
    def addVertex(self):
        self.vertices = np.arange(len(self.vertices) + 1)
            
    def rmEdge(self, v1, v2):
        if type(v1) != int or type(v2) != int:
            raise vertexError("Vertices are integers")
        else:
            if [v1, v2] in self.edges:
                self.edges.remove([v1, v2])
            elif [v2, v1] in self.edges:
                self.edges.remove([v2, v1])
            else:
                return('Edge not in edge set')
            
    #This function broke my indexing for other functions and it was too much hassle to alter the indexes
#     def rmVertex(self, vertex):
#         if vertex in self.vertices:
#             self.vertices.remove(vertex)
#             for edge in self.edges:
#                 if vertex in edge:
#                     self.edges.remove(edge)
#         else:
#             print('Not in vertex set')
    
    def printGraph(self):
        print(f'Vertex set: {self.vertices}\nEdge set: {self.edges}')
        
    def adjMatrix(self): # Gives the adjacency matrix of the graph. Used in other functions
        #Make an n by n zero matrix
        matrix = np.zeros(len(self.vertices)*len(self.vertices)).reshape(len(self.vertices), len(self.vertices))
        for edge in self.edges:
            matrix[edge[0]][edge[1]] = 1 #Graph is undirected and unweighted, so both transposes are set to 1
            matrix[edge[1]][edge[0]] = 1
        return matrix
        
        
    # I got some help from here https://www.geeksforgeeks.org/implementation-of-dfs-using-adjacency-matrix/
    # because my dumb brain can't handle recursion
    
    # I had visited = {} as default and all my code was broken until I 
    # changed it to None. Python is confusing!
    
    # showSteps parameter allows me to not print the DFS steps when the function is called by hasCycles, isConnected, kruskal
    def DFS(self, start, visited = None, previous = None, showSteps = True, output = None):
        if start in self.vertices:
            toPrint = showSteps # Need to pass this in to make it work with recursion
            verts = self.vertices
            hasCycles = False #Used in the hasCycles method
            matrix = self.adjMatrix() # I could only figure out how to make this work for adjacency matrices
            if showSteps == True: #We want to block print when DFS called from other functions
                print(f'Visit {start} from {previous}')
            if visited == None:
                visited = {start}
                output = [start]
            else:
                visited.add(start)
                output.append(start)
            for i in verts:
                if matrix[start][i] != 0:
                    if i in visited and i != previous:
                        self.__cycles = True
                    elif i not in visited:
                        self.DFS(i, visited, previous = start, showSteps = toPrint, output = output)
            self.__connected = visited #We need these in the following two functions
            return output
        else: return 'Invalid starting vertex'
        
    def isConnected(self):
        self.DFS(0, showSteps= False)
        # Start doesn't matter, since we can reach any other vertex iff connected. Choose 0 since its in any size graph
        if len(self.__connected) == len(self.vertices):
            return True
        else:
            return False
    
    def hasCycles(self):
        for i in self.vertices:
            self.DFS(i, showSteps = False)
        return self.__cycles



#Vertices and edges as above. weights must be a list of numbers of same length as edge set, and applies 
#Weights to edges in index-wise fashion
import numpy as np
import pandas as pd
import copy

import graphsTrees.graphs.graphs as gr
from graphsTrees.exceptions.EricExceptions import weightError
from graphsTrees.exceptions.EricExceptions import vertexError
from graphsTrees.exceptions.EricExceptions import edgeError

class wtGraph(gr.Graph):
    def __init__(self, vertices, edges, weights):
        gr.Graph.__init__(self, vertices, edges)
        self.weights = weights
                #Adding this condition breaks the kruskal algorithm 
#         if len(edges) != len(weights):
#             raise IndexError("Edge and weight sets must be same length")
        for i in weights:
            try:
                float(i)
            except:
                raise weightError('Weights must be numbers')
                
    #Alter addEdge method to accept a weight argument
    def addEdge(self, v1, v2, weight):
        try:
            int(weight)
        except ValueError:
            raise weightError("Weight must be a number")
        else:
            if type(v1) != int or type(v2) != int:
                raise vertexError('Vertices are integers')
            elif (v1 in self.vertices and v2 in self.vertices):
                if [v1, v2] not in self.edges and [v2, v1] not in self.edges:
                    self.edges.append([v1, v2])
                    self.weights.append(weight)
                else:
                    return('Edge already in edge set')
            else:
                return('Input valid vertices')
            
     #Alter method to remove weight as well
    def rmEdge(self, v1, v2):
        if type(v1) != int or type(v2) != int:
            raise vertexError("Vertices are integers")
        else:
            if [v1, v2] in self.edges:
                self.weights.remove(self.weights[self.edges.index([v1,v2])])
                self.edges.remove([v1, v2])
            elif [v2, v1] in self.edges:
                self.weights.remove(self.weights[self.edges.index([v2,v1])])
                self.edges.remove([v2, v1])
            else:
                return('Edge not in edge set')
            
    #Alter method to print weights as well        
    def printGraph(self):
        return(f'Vertex set: {self.vertices}\nEdge set: {self.edges}\nWeights:{self.weights}')
        
    #Adjust the adjMatrix function to incorperate weights
    def adjMatrix(self): # Gives the adjacency matrix of the graph. Used in other functions
        #Make an n by n zero matrix
        matrix = np.zeros(len(self.vertices)*len(self.vertices)).reshape(len(self.vertices), len(self.vertices))
        for edge in self.edges:
            matrix[edge[0]][edge[1]] = self.weights[self.edges.index(edge)] #Graph is undirected and weighted
            matrix[edge[1]][edge[0]] = self.weights[self.edges.index(edge)]
        return matrix
    

    def kruskal(self):
        kruskVertices = copy.copy(self.vertices)
        # Python pointers strike again! Code was broken for like a day before I remembered copy.copy
        kruskEdges = copy.copy(self.edges)
        kruskWeights = copy.copy(self.weights)
        tree = []
        if self.isConnected() == True:
            while len(tree) < len(kruskVertices) - 2:
                #This is a horrible way of  using the index from the weights list in the edges list
                #If adding the edge makes a cycle, just remove it. else, add to tree
                if wtGraph(len(kruskVertices), tree + [kruskEdges[kruskWeights.index(min(kruskWeights))]], kruskWeights).hasCycles():
                    kruskEdges.remove(kruskEdges[kruskWeights.index(min(kruskWeights))])
                    kruskWeights.remove(min(kruskWeights))
                else:
                    tree.append(kruskEdges[kruskWeights.index(min(kruskWeights))])
                    kruskEdges.remove(kruskEdges[kruskWeights.index(min(kruskWeights))])
                    kruskWeights.remove(min(kruskWeights))
            #This adds the last edge
            tree.append(kruskEdges[kruskWeights.index(min(kruskWeights))])
            return tree
        else:
            return('Graph is disconnected, and there is no spanning tree')
    

    def totalWeight(self):
        return sum(self.weights)


from graphsTrees.trees.node import Node
from graphsTrees.trees.tree import Tree
from graphsTrees.exceptions.TreeExceptions import NodeError
from graphsTrees.exceptions.TreeExceptions import NodeKeyError

class BST(Tree):
    
    def create_node(self,key):
        try: 
            return Node(int(key))
        except ValueError:
            raise NodeKeyError
    
    def insert(self, node, key):
        if node is None:
            return self.create_node(key)
        if key < node.key:
            node.left = self.insert(node.left, key)
        elif key > node.key:
            node.right = self.insert(node.right, key)
        return node
    
    def search_node(self,node,key):
        try: 
            if node is None:
                return "Node {} does not exist".format(int(key))
            if node.key == int(key):
                return "Node {} exists".format(int(key))
            if node.key<int(key):
                return self.search_node(node.right,int(key))
            else:
                return self.search_node(node.left,int(key))
        except ValueError:
            raise NodeKeyError
    
    def delete_node(self,node,key):
        try: 
            if int(key) < node.key:
                node.left = self.delete_node(node.left,int(key))
            elif int(key)>node.key:
                node.right = self.delete_node(node.right,int(key))
            else:
                if node.left is None and node.right is None:
                    return None
                elif node.left is None:
                    swap = node.right
                    del node
                    return swap 
                elif node.right is None:
                    swap = node.left
                    del node
                    return swap
            return node
        except ValueError:
            raise NodeKeyError





from setuptools import setup, find_packages

setup(
name='lab4GraphsTrees',
version='0.1',
packages=find_packages(exclude=['tests*']),
license='MIT',
description='An assignment package for graph and tree data structures',
url='https://github.com/adityasal/DATA533_Lab4',
author='Eric Baxter, Aditya Saluja',
author_email='ebaxter1@hotmail.ca'
)